@@ -12,6 +12,7 @@ import sys
 import time
 import traceback
 import random
+import re
 import mozinfo
 import moznetwork
 import xml.dom.minidom as dom
@@ -214,6 +215,7 @@ class MarionetteTestResult(unittest._TextTestResult, TestResultCollection):
                 self.stream.writeln('END LOG:')
 
     def printErrorList(self, flavour, errors):
+        TIMEOUT_MESSAGE = "ScriptTimeoutException: ScriptTimeoutException: timed out"
         for error in errors:
             err = error.output
             self.stream.writeln(self.separator1)
@@ -221,17 +223,23 @@ class MarionetteTestResult(unittest._TextTestResult, TestResultCollection):
             self.stream.writeln(self.separator2)
             lastline = None
             fail_present = None
+            test_name = self.getInfo(error)
             for line in err:
-                if not line.startswith('\t'):
+                if not line.startswith('\t') and line != '':
                     lastline = line
                 if 'TEST-UNEXPECTED-FAIL' in line:
                     fail_present = True
             for line in err:
                 if line != lastline or fail_present:
-                    self.stream.writeln("%s" % line)
+                    if re.match('.*\.js', test_name):
+                        if error.reason != TIMEOUT_MESSAGE:
+                            self.stream.writeln("%s" % line)
+                    else:
+                        self.stream.writeln("%s" % line)
+
                 else:
                     self.stream.writeln("TEST-UNEXPECTED-FAIL | %s | %s" %
-                                        (self.getInfo(error), line))
+                                        (test_name, error.reason))
 
     def stopTest(self, *args, **kwargs):
         unittest._TextTestResult.stopTest(self, *args, **kwargs)
@@ -251,7 +259,7 @@ class MarionetteTextTestRunner(unittest.TextTestRunner):
         self.b2g_pid = None
         del kwargs['marionette']
 
-        if self.capabilities['device'] != 'desktop' and self.capabilities['b2g']:
+        if self.capabilities["device"] != "desktop" and self.capabilities["browserName"] == "B2G":
             def b2g_pre_run():
                 dm_type = os.environ.get('DM_TRANS', 'adb')
                 if dm_type == 'adb':
@@ -463,6 +471,29 @@ class BaseMarionetteOptions(OptionParser):
                         dest='shuffle',
                         default=False,
                         help='run tests in a random order')
+        self.add_option('--total-chunks',
+                        dest='total_chunks',
+                        type=int,
+                        help='how many chunks to split the tests up into')
+        self.add_option('--this-chunk',
+                        dest='this_chunk',
+                        type=int,
+                        help='which chunk to run')
+        self.add_option('--sources',
+                        dest='sources',
+                        action='store',
+                        help='path to sources.xml (Firefox OS only)')
+        self.add_option('--server-root',
+                        dest='server_root',
+                        action='store',
+                        help='sets the web server\'s root directory to the given path')
+        self.add_option('--gecko-log',
+                        dest='gecko_log',
+                        action='store',
+                        help="Define the path to store log file. If the path is"
+                             " a directory, the real log file will be created"
+                             " given the format gecko-(timestamp).log. If it is"
+                             " a file, if will be used directly. Default: 'gecko.log'")
 
     def parse_args(self, args=None, values=None):
         options, tests = OptionParser.parse_args(self, args, values)
@@ -480,6 +511,10 @@ class BaseMarionetteOptions(OptionParser):
             print 'must specify --binary, --emulator or --address'
             sys.exit(1)
 
+        if options.emulator and options.bin:
+            print 'can\'t specify both --emulator and --binary'
+            sys.exit(1)
+
         if not options.es_servers:
             options.es_servers = ['elasticsearch-zlb.dev.vlan81.phx.mozilla.com:9200',
                                   'elasticsearch-zlb.webapp.scl3.mozilla.com:9200']
@@ -500,6 +535,18 @@ class BaseMarionetteOptions(OptionParser):
             raise ValueError('Invalid emulator resolution format. '
                              'Should be like "480x800".')
 
+        if options.total_chunks is not None and options.this_chunk is None:
+            self.error('You must specify which chunk to run.')
+
+        if options.this_chunk is not None and options.total_chunks is None:
+            self.error('You must specify how many chunks to split the tests into.')
+
+        if options.total_chunks is not None:
+            if not 1 <= options.total_chunks:
+                self.error('Total chunks must be greater than 1.')
+            if not 1 <= options.this_chunk <= options.total_chunks:
+                self.error('Chunk to run must be between 1 and %s.' % options.total_chunks)
+
         for handler in self.verify_usage_handlers:
             handler(options, tests)
 
@@ -517,7 +564,9 @@ class BaseMarionetteTestRunner(object):
                  logcat_dir=None, xml_output=None, repeat=0, gecko_path=None,
                  testvars=None, tree=None, type=None, device_serial=None,
                  symbols_path=None, timeout=None, es_servers=None, shuffle=False,
-                 sdcard=None, **kwargs):
+                 sdcard=None, this_chunk=1, total_chunks=1, sources=None, server_root=None,
+                 gecko_log=None,
+                 **kwargs):
         self.address = address
         self.emulator = emulator
         self.emulatorBinary = emulatorBinary
@@ -534,7 +583,6 @@ class BaseMarionetteTestRunner(object):
         self.logger = logger
         self.noWindow = noWindow
         self.httpd = None
-        self.baseurl = None
         self.marionette = None
         self.logcat_dir = logcat_dir
         self.xml_output = xml_output
@@ -553,6 +601,11 @@ class BaseMarionetteTestRunner(object):
         self.es_servers = es_servers
         self.shuffle = shuffle
         self.sdcard = sdcard
+        self.sources = sources
+        self.server_root = server_root
+        self.this_chunk = this_chunk
+        self.total_chunks = total_chunks
+        self.gecko_log = gecko_log
         self.mixin_run_tests = []
         self.manifest_skipped_tests = []
         self.tests = []
@@ -587,7 +640,7 @@ class BaseMarionetteTestRunner(object):
         # for XML output
         self.testvars['xml_output'] = self.xml_output
         self.results = []
- 
+
     @property
     def capabilities(self):
         if self._capabilities:
@@ -617,78 +670,80 @@ class BaseMarionetteTestRunner(object):
     def reset_test_stats(self):
         self.passed = 0
         self.failed = 0
+        self.unexpected_successes = 0
         self.todo = 0
+        self.skipped = 0
         self.failures = []
 
-    def start_httpd(self):
-        host = moznetwork.get_ip()
+    def start_httpd(self, need_external_ip):
+        host = "127.0.0.1"
+        if need_external_ip:
+            host = moznetwork.get_ip()
+        docroot = self.server_root or os.path.join(os.path.dirname(os.path.dirname(__file__)), 'www')
+        if not os.path.isdir(docroot):
+            raise Exception("Server root %s is not a valid path" % docroot)
         self.httpd = MozHttpd(host=host,
                               port=0,
-                              docroot=os.path.join(os.path.dirname(os.path.dirname(__file__)), 'www'))
+                              docroot=docroot)
         self.httpd.start()
-        self.baseurl = 'http://%s:%d/' % (host, self.httpd.httpd.server_port)
-        self.logger.info('running webserver on %s' % self.baseurl)
+        self.marionette.baseurl = 'http://%s:%d/' % (host, self.httpd.httpd.server_port)
+        self.logger.info('running webserver on %s' % self.marionette.baseurl)
+
+
+    def _build_kwargs(self):
+        kwargs = {
+            'device_serial': self.device_serial,
+            'symbols_path': self.symbols_path,
+            'timeout': self.timeout,
+        }
+        if self.bin:
+            kwargs.update({
+                'host': 'localhost',
+                'port': 2828,
+                'app': self.app,
+                'app_args': self.app_args,
+                'bin': self.bin,
+                'profile': self.profile,
+                'gecko_log': self.gecko_log,
+            })
+
+        if self.emulator:
+            kwargs.update({
+                'homedir': self.homedir,
+                'logcat_dir': self.logcat_dir,
+                'gecko_path': self.gecko_path,
+            })
+
+        if self.address:
+            host, port = self.address.split(':')
+            kwargs.update({
+                'host': host,
+                'port': int(port),
+            })
+            if self.emulator:
+                kwargs['connectToRunningEmulator'] = True
+
+            if not self.bin:
+                try:
+                    #establish a socket connection so we can vertify the data come back
+                    connection = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
+                    connection.connect((host,int(port)))
+                    connection.close()
+                except Exception, e:
+                    raise Exception("Connection attempt to %s:%s failed with error: %s" %(host,port,e))
+        elif self.emulator:
+            kwargs.update({
+                'emulator': self.emulator,
+                'emulatorBinary': self.emulatorBinary,
+                'emulatorImg': self.emulatorImg,
+                'emulator_res': self.emulator_res,
+                'noWindow': self.noWindow,
+                'sdcard': self.sdcard,
+            })
+        return kwargs
 
     def start_marionette(self):
-        assert(self.baseurl is not None)
-        if self.bin:
-            if self.address:
-                host, port = self.address.split(':')
-            else:
-                host = 'localhost'
-                port = 2828
-            self.marionette = Marionette(host=host,
-                                         port=int(port),
-                                         app=self.app,
-                                         app_args=self.app_args,
-                                         bin=self.bin,
-                                         profile=self.profile,
-                                         baseurl=self.baseurl,
-                                         timeout=self.timeout,
-                                         device_serial=self.device_serial)
-        elif self.address:
-            host, port = self.address.split(':')
-            try:
-                #establish a socket connection so we can vertify the data come back
-                connection = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
-                connection.connect((host,int(port)))
-                connection.close()
-            except Exception, e:
-                raise Exception("Could not connect to given marionette host:port: %s" % e)
-            if self.emulator:
-                self.marionette = Marionette.getMarionetteOrExit(
-                                             host=host, port=int(port),
-                                             connectToRunningEmulator=True,
-                                             homedir=self.homedir,
-                                             baseurl=self.baseurl,
-                                             logcat_dir=self.logcat_dir,
-                                             gecko_path=self.gecko_path,
-                                             symbols_path=self.symbols_path,
-                                             timeout=self.timeout,
-                                             device_serial=self.device_serial)
-            else:
-                self.marionette = Marionette(host=host,
-                                             port=int(port),
-                                             baseurl=self.baseurl,
-                                             timeout=self.timeout,
-                                             device_serial=self.device_serial)
-        elif self.emulator:
-            self.marionette = Marionette.getMarionetteOrExit(
-                                         emulator=self.emulator,
-                                         emulatorBinary=self.emulatorBinary,
-                                         emulatorImg=self.emulatorImg,
-                                         emulator_res=self.emulator_res,
-                                         homedir=self.homedir,
-                                         baseurl=self.baseurl,
-                                         noWindow=self.noWindow,
-                                         logcat_dir=self.logcat_dir,
-                                         gecko_path=self.gecko_path,
-                                         symbols_path=self.symbols_path,
-                                         timeout=self.timeout,
-                                         sdcard=self.sdcard,
-                                         device_serial=self.device_serial)
-        else:
-            raise Exception("must specify binary, address or emulator")
+        self.marionette = Marionette(**self._build_kwargs())
 
     def post_to_autolog(self, elapsedtime):
         self.logger.info('posting results to autolog')
@@ -738,10 +793,7 @@ class BaseMarionetteTestRunner(object):
         self.reset_test_stats()
         starttime = datetime.utcnow()
 
-        if not self.httpd:
-            print "starting httpd"
-            self.start_httpd()
-
+        need_external_ip = True
         if not self.marionette:
             self.start_marionette()
             if self.emulator:
@@ -749,6 +801,14 @@ class BaseMarionetteTestRunner(object):
             # Retrieve capabilities for later use
             if not self._capabilities:
                 self.capabilities
+            # if we're working against a desktop version, we usually don't need
+            # an external ip
+            if self._capabilities['device'] == "desktop":
+                need_external_ip = False
+
+        if not self.httpd:
+            print "starting httpd"
+            self.start_httpd(need_external_ip)
 
         for test in tests:
             self.add_test(test)
@@ -762,8 +822,14 @@ class BaseMarionetteTestRunner(object):
             counter -= 1
         self.logger.info('\nSUMMARY\n-------')
         self.logger.info('passed: %d' % self.passed)
-        self.logger.info('failed: %d' % self.failed)
-        self.logger.info('todo: %d' % self.todo)
+        if self.unexpected_successes == 0:
+            self.logger.info('failed: %d' % self.failed)
+        else:
+            self.logger.info('failed: %d (unexpected sucesses: %d)', self.failed, self.unexpected_successes)
+        if self.skipped == 0:
+            self.logger.info('todo: %d', self.todo)
+        else:
+            self.logger.info('todo: %d (skipped: %d)', self.todo, self.skipped)
 
         if self.failed > 0:
             self.logger.info('\nFAILED TESTS\n-------')
@@ -930,16 +996,18 @@ class BaseMarionetteTestRunner(object):
 
             self.failed += len(results.failures) + len(results.errors)
             if hasattr(results, 'skipped'):
+                self.skipped += len(results.skipped)
                 self.todo += len(results.skipped)
             self.passed += results.passed
             for failure in results.failures + results.errors:
                 self.failures.append((results.getInfo(failure), failure.output, 'TEST-UNEXPECTED-FAIL'))
             if hasattr(results, 'unexpectedSuccesses'):
                 self.failed += len(results.unexpectedSuccesses)
+                self.unexpected_successes += len(results.unexpectedSuccesses)
                 for failure in results.unexpectedSuccesses:
                     self.failures.append((results.getInfo(failure), 'TEST-UNEXPECTED-PASS'))
             if hasattr(results, 'expectedFailures'):
-                self.passed += len(results.expectedFailures)
+                self.todo += len(results.expectedFailures)
 
     def run_test_set(self, tests):
         if self.shuffle:
@@ -951,6 +1019,20 @@ class BaseMarionetteTestRunner(object):
                 break
 
     def run_test_sets(self):
+        if self.total_chunks > len(self.tests):
+            raise ValueError('Total number of chunks must be between 1 and %d.' % len(self.tests))
+        if self.total_chunks > 1:
+            chunks = [[] for i in range(self.total_chunks)]
+            for i, test in enumerate(self.tests):
+                target_chunk = i % self.total_chunks
+                chunks[target_chunk].append(test)
+
+            self.logger.info('Running chunk %d of %d (%d tests selected from a '
+                             'total of %d)' % (self.this_chunk, self.total_chunks,
+                                               len(chunks[self.this_chunk - 1]),
+                                               len(self.tests)))
+            self.tests = chunks[self.this_chunk - 1]
+
         oop_tests = [x for x in self.tests if x.get('oop')]
         self.run_test_set(oop_tests)
 
